from textwrap import dedent

from django.shortcuts import get_object_or_404, redirect
from drf_spectacular.utils import extend_schema, extend_schema_view
from rest_framework.authentication import SessionAuthentication, TokenAuthentication
from rest_framework.generics import ListCreateAPIView, RetrieveUpdateAPIView, UpdateAPIView
from rest_framework.views import APIView

from arkindex.ponos.models import Artifact, Task
from arkindex.ponos.permissions import (
    IsAgentOrArtifactGuest,
    IsAgentOrTaskGuest,
    IsAssignedAgentOrReadOnly,
    IsAssignedAgentOrTaskOrReadOnly,
    IsTaskAdmin,
)
from arkindex.ponos.serializers import ArtifactSerializer, TaskSerializer, TaskTinySerializer


@extend_schema(tags=["ponos"])
@extend_schema_view(
    get=extend_schema(
        operation_id="RetrieveTaskFromAgent",
        description=dedent("""
        Retrieve a Ponos task.

        Requires **guest** access on the task's process, Ponos agent authentication, or authentication as the Ponos task to retrieve.
        """),
    ),
    put=extend_schema(
        operation_id="UpdateTaskFromAgent",
        description=dedent("""
        Update a task.

        Requires authentication as the Ponos agent assigned to the task.
        """),
    ),
    patch=extend_schema(
        operation_id="PartialUpdateTaskFromAgent",
        description=dedent("""
        Partially update a task.

        Requires authentication as the Ponos agent assigned to the task.
        """),
    ),
)
class TaskDetailsFromAgent(RetrieveUpdateAPIView):

    # Avoid stale read when a recently assigned agent wants to update
    # the state of one of its tasks
    queryset = Task.objects.using("default").select_related(
        "agent__farm",
        "gpu",
        # Used for permission checks
        "process__corpus",
        "process__revision__repo",
    )
    permission_classes = (
        # On all HTTP methods, require either any Ponos agent, an instance admin, the task itself, or guest access to the process' task
        IsAgentOrTaskGuest,
        # On unsafe HTTP methods, require a Ponos agent assigned to the task. Both permission classes are combined.
        IsAssignedAgentOrReadOnly,
    )
    serializer_class = TaskSerializer


@extend_schema(tags=["ponos"])
@extend_schema_view(
    get=extend_schema(
        operation_id="ListArtifacts",
        description=dedent("""
        List the artifacts of a task.

        Requires **guest** access on the task's process, Ponos agent authentication, or authentication as the Ponos task to list artifacts of.
        """),
    ),
    post=extend_schema(
        operation_id="CreateArtifact",
        description=dedent("""
        Create an artifact on task.

        Requires authentication as the Ponos task to create an artifact on, or as a Ponos agent assigned to the task.
        """),
    ),
)
class TaskArtifacts(ListCreateAPIView):

    # Used for OpenAPI schema serialization: the ID in the path is the task ID
    queryset = Task.objects.none()
    serializer_class = ArtifactSerializer
    permission_classes = (
        # On all HTTP methods, require either any Ponos agent, an instance admin, the task itself, or guest access to the process' task
        IsAgentOrTaskGuest,
        # On unsafe HTTP methods, require a Ponos agent assigned to the task, or the task itself. Both permission classes are combined.
        IsAssignedAgentOrTaskOrReadOnly,
    )

    # Force no pagination, even when global settings add them
    pagination_class = None

    @property
    def task(self):
        task = get_object_or_404(
            # Select the required tables for permissions checking
            Task.objects.select_related("process__corpus", "process__revision__repo"),
            pk=self.kwargs["pk"],
        )
        self.check_object_permissions(self.request, task)
        return task

    def get_queryset(self):
        return self.task.artifacts.all()

    def perform_create(self, serializer):
        # Assign task when creating through the API
        serializer.save(task=self.task)


class TaskArtifactDownload(APIView):
    """
    Redirects to the S3 URL of an artifact in order to download it.
    """

    permission_classes = (IsAgentOrArtifactGuest, )

    def get_object(self, pk, path):
        artifact = get_object_or_404(
            # Select the required tables for permissions checking
            Artifact.objects.select_related("task__process__corpus", "task__process__revision__repo"),
            task_id=pk,
            path=path,
        )
        self.check_object_permissions(self.request, artifact)
        return artifact

    @extend_schema(
        operation_id="DownloadArtifact",
        responses={302: None},
        tags=["ponos"],
    )
    def get(self, request, *args, **kwargs):
        artifact = self.get_object(*args, **kwargs)
        return redirect(artifact.s3_url)


@extend_schema(tags=["ponos"])
@extend_schema_view(
    put=extend_schema(
        description=dedent("""
        Update a task.

        Requires **admin** access on the task's process, or to be the creator of the process.
        Cannot be used with Ponos agent or task authentication.
        """),
    ),
    patch=extend_schema(
        description=dedent("""
        Partially update a task.

        Requires **admin** access on the task's process, or to be the creator of the process.
        Cannot be used with Ponos agent or task authentication.
        """),
    ),
)
class TaskUpdate(UpdateAPIView):
    # Only allow regular users, not Ponos agents or tasks
    authentication_classes = (TokenAuthentication, SessionAuthentication)
    # Only allow regular users that have admin access to the task's process
    permission_classes = (IsTaskAdmin, )
    queryset = Task.objects.select_related("process__corpus", "process__revision__repo")
    serializer_class = TaskTinySerializer
