from django.urls import reverse
from rest_framework import status
from rest_framework.serializers import DateTimeField

from arkindex.ponos.models import Farm
from arkindex.process.models import ProcessMode, Repository
from arkindex.project.tests import FixtureTestCase
from arkindex.users.models import Role, User


class TestRepositories(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.repo = Repository.objects.get(url="http://my_repo.fake/workers/worker")
        cls.repo_2 = Repository.objects.get(url="http://gitlab/repo")
        cls.rev = cls.repo.revisions.get()
        cls.serialized_revision = {
            "id": str(cls.rev.id),
            "repo_id": str(cls.repo.id),
            "author": cls.rev.author,
            "commit_url": cls.rev.commit_url,
            "date": DateTimeField().to_representation(value=cls.rev.created),
            "hash": cls.rev.hash,
            "message": cls.rev.message
        }
        cls.farm = Farm.objects.get()

    def build_repository_response(self):
        return [
            {
                "id": str(self.repo_2.id),
                "url": self.repo_2.url,
                "workers": [],
                "authorized_users": self.repo_2.memberships.count(),
            }, {
                "id": str(self.repo.id),
                "url": self.repo.url,
                "workers": [
                    {
                        "id": str(w.id),
                        "name": w.name,
                        "type": w.type.slug,
                        "slug": w.slug,
                    } for w in self.repo.workers.all()
                ],
                "authorized_users": self.repo.memberships.count(),
            }
        ]

    def test_list_repository_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:repository-list"))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list_repository_user_readable(self):
        self.repo.memberships.create(user=self.user, level=Role.Admin.value)
        self.repo_2.memberships.create(user=self.user, level=Role.Guest.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(reverse("api:repository-list"))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data["results"]), 2)
        self.assertCountEqual(data["results"], self.build_repository_response())

    def test_list_repository_multiple_refs(self):
        """
        Listed repositories should not be duplicated
        """
        self.client.force_login(self.superuser)
        with self.assertNumQueries(7):
            response = self.client.get(reverse("api:repository-list"))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data["results"]), 2)
        self.assertCountEqual(data["results"], self.build_repository_response())

    def test_repository_retrieve_requires_login(self):
        response = self.client.get(
            reverse("api:repository-retrieve", kwargs={"pk": str(self.repo.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_repository_retrieve_wrong_id(self):
        self.client.force_login(self.user)
        response = self.client.get(
            reverse("api:repository-retrieve", kwargs={"pk": "12341234-1234-1234-1234-123412341234"})
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.json(), {"detail": "Not found."})

    def test_repository_retrieve(self):
        """
        An user with an access right on a repository is able to retrieve it without its clone URL
        """
        self.repo_2.memberships.create(user=self.user, level=Role.Guest.value)
        self.client.force_login(self.user)
        response = self.client.get(
            reverse("api:repository-retrieve", kwargs={"pk": str(self.repo_2.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertDictEqual(data, {
            "id": str(self.repo_2.id),
            "url": self.repo_2.url,
            "workers": [],
            "authorized_users": self.repo_2.memberships.count(),
        })

    def test_repository_retrieve_ponos_task(self):
        """
        A Ponos task can retrieve the clone URL for the repository it should work on
        """
        process = self.repo_2.revisions.first().processes.create(
            mode=ProcessMode.Repository,
            creator=self.user,
            farm=Farm.objects.first(),
        )
        process.run()
        task = process.tasks.get()
        self.repo_2.memberships.create(user=self.user, level=Role.Guest.value)

        with self.assertNumQueries(3):
            response = self.client.get(
                reverse("api:repository-retrieve", kwargs={"pk": str(self.repo_2.id)}),
                HTTP_AUTHORIZATION=f"Ponos {task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": str(self.repo_2.id),
            "url": self.repo_2.url,
            "workers": [],
            "authorized_users": self.repo_2.memberships.count(),
        })

    def test_repository_retrieve_ponos_task_not_assigned(self):
        """
        A Ponos task cannot retrieve the clone URL of a repository it should not work on
        """
        process = self.repo.revisions.first().processes.create(
            mode=ProcessMode.Repository,
            creator=self.user,
            farm=Farm.objects.first(),
        )
        process.run()
        task = process.tasks.get()
        self.repo_2.memberships.create(user=self.user, level=Role.Guest.value)

        with self.assertNumQueries(3):
            response = self.client.get(
                reverse("api:repository-retrieve", kwargs={"pk": str(self.repo_2.id)}),
                HTTP_AUTHORIZATION=f"Ponos {task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            "id": str(self.repo_2.id),
            "url": self.repo_2.url,
            "workers": [],
            "authorized_users": self.repo_2.memberships.count(),
        })

    def test_repository_retrieve_disabled_repo(self):
        self.repo_2.memberships.create(user=self.user, level=Role.Guest.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.get(reverse("api:repository-retrieve", kwargs={"pk": str(self.repo_2.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertDictEqual(data, {
            "id": str(self.repo_2.id),
            "url": self.repo_2.url,
            "workers": [],
            "authorized_users": self.repo_2.memberships.count(),
        })

    def test_repository_delete_not_admin(self):
        """
        A user cannot delete a repository if he has no admin access
        """
        creator = User.objects.create()
        repo = Repository.objects.create(url="http://somewhere.com/repo")
        repo.memberships.create(user=creator, level=Role.Contributor.value)
        self.client.force_login(creator)
        response = self.client.delete(reverse("api:repository-retrieve", kwargs={"pk": str(repo.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_repository_delete_creator_write_access(self):
        """
        An admin member of a repository can delete it
        """
        self.repo.memberships.create(user=self.user, level=Role.Admin.value)
        self.client.force_login(self.user)
        response = self.client.delete(reverse("api:repository-retrieve", kwargs={"pk": str(self.repo.id)}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        with self.assertRaises(Repository.DoesNotExist):
            self.repo.refresh_from_db()

    def test_revision_retrieve_requires_login(self):
        response = self.client.get(
            reverse("api:revision-retrieve", kwargs={"pk": str(self.rev.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_revision_retrieve_wrong_id(self):
        self.client.force_login(self.user)
        response = self.client.get(
            reverse("api:revision-retrieve", kwargs={"pk": "12341234-1234-1234-1234-123412341234"})
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.json(), {"detail": "Not found."})

    def test_revision_retrieve_contributor_right(self):
        """
        A user with a contributor access to the revision repo is allowed to retrieve information
        """
        self.repo.memberships.create(user=self.user, level=Role.Contributor.value)
        self.client.force_login(self.user)
        response = self.client.get(
            reverse("api:revision-retrieve", kwargs={"pk": str(self.rev.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertDictEqual(data, self.serialized_revision)
